<?php
namespace WebFacing\cPanel\Email;

/**
 * Exit if accessed directly
 */
if ( ! \class_exists( '\WP' ) ) {
	exit;
}

class Box_Table extends \WP_List_Table {

	/**
	 * Prepare the items for the table to process
	 *
	 * @return never
	 */
	public function prepare_items() {

		$this->process_bulk_action();

		$columns  = $this->get_columns();
		$hidden   = $this->get_hidden_columns();
		$sortable = $this->get_sortable_columns();
		$data	  = $this->table_data();
		$cur_page = $this->get_pagenum();
		$per_page = 50;

		\usort( $data, [ $this, 'sort_data' ] );
		$total_items  = \count( $data );

		$this->set_pagination_args( [
			'total_items' => $total_items,
			'per_page'    => $per_page
		] );

		$data = \array_slice( $data, ( $cur_page - 1 ) * $per_page, $per_page );

		$this->_column_headers = [ $columns, $hidden, $sortable ];
		$this->items = $data;
	}

	/**
	 * Override the parent columns method. Defines the columns to use in your listing table
	 *
	 * @return array
	 */
	public function get_columns(): array {
		$columns = [
			'cb'           => '<input type="checkbox" />',
			'email'	       => \_x( 'Email', 'Column Header', Plugin::$text_domain ),
			'box_guid'     => 'GUID',
			'box_name'     => \_x( 'Mailbox', 'Column Header', Plugin::$text_domain ),
			'messages'     => \_x( 'Messages', 'Column Header', Plugin::$text_domain ),
			'diskused'     => \_x( 'Used', 'Column Header', Plugin::$text_domain ),
			'diskused_num' => 'DUN',
		];

		return $columns;
	}

	/**
	 * Define which columns are hidden
	 *
	 * @return array
	 */
	public function get_hidden_columns(): array {
		return [ 'email', 'box_guid', 'diskused_num' ];
	}

	/**
	 * Define the sortable columns
	 *
	 * @return array
	 */
	public function get_sortable_columns(): array {
		return [
			'box_name'  => [ 'box_name', false ],
			'messages'  => [ 'massages', false ],
			'diskused'  => [ 'diskused_num', false ],
		];
	}

	/**
	 * Get the table data
	 *
	 * @return array
	 */
	private function table_data(): array {
		$data = [];
		$email = UAPI::main_email_account()->email === $this->_args['email'] ? '' : $this->_args['email'];

		foreach ( UAPI::email_mailboxes( $email ) as $box ) {
			$data[] = [
//				'email'        => $this->_args['email'],
				'box_guid'     => $box->guid,
				'box_name'     => $box->mailbox,
				'messages'     => \intval( $box->messages ),
				'diskused'     => \size_format( $box->vsize ),
				'diskused_num' => $box->vsize,
			];
		}
		return $data;
	}

	/**
	 * Render the bulk edit checkbox
	 *
	 * @param array $item
	 *
	 * @return string
	 */
	public function column_cb( $item ) {
		return $item['messages'] ?
			\sprintf(
				'<input type="checkbox" name="guid[]" value="%1$s" />',
				\esc_attr( $item['box_guid'] ),
			) :
		'';
	}

	/**
	 * Render the bulk edit checkbox
	 *
	 * @param array $item
	 *
	 * @return string
	 */
	public function column_messages( $item ) {
		$page = \esc_attr( $_REQUEST['page'] );
		$actions = $item['messages'] ? [
			'row-delete'   => '<a href="' . \add_query_arg( [
				'page'     => $page,
				'email'    => $this->_args['email'],
				'box'      => $item['box_name'],
				'guid'     => $item['box_guid'],
				'action'   => 'row-delete',
				'_wpnonce' => \wp_create_nonce( 'row-delete' ),
			], \admin_url( 'admin.php' ) ) . '">' . _x( 'Delete messages', 'List Table Row Action', Plugin::$text_domain ),
		] :
		[];
		return $item[ 'messages' ] . '<br />' . $this->row_actions( $actions );
	}

	/**
	 * Define what data to show on each column of the table
	 *
	 * @param  array  $item	       - Data
	 * @param  string $column_name - Current column name
	 *
	 * @return mixed
	 */
	public function column_default( $item, $column_name ) {
		switch ( $column_name ) {
			case 'box_name':
				return '<span title="Native name: ' . $item[ $column_name ] . '.">' . self::display_mailbox( $item[ $column_name ] ) . '</span>';
			case 'messages':
			case 'diskused':
				return $item[ $column_name ];
			default:
				return \print_r( $item, true ) ;
		}
	}

	/**
	 * Allows you to sort the data by the variables set in the $_GET
	 *
	 * @return mixed
	 */
	private function sort_data( $a, $b ) {
		// Set defaults
		$orderby  = 'box_name';
		$order    = 'asc';

		// If orderby is set, use this as the sort column
		$orderby  = \sanitize_text_field( $_GET['orderby'] ?? $orderby );
//		$orderby  = $orderby === 'diskused' ? 'diskused_num' : $orderby;

		// If order is set use this as the order
		$order    = \sanitize_text_field( $_GET['order'] ?? $order );

		$result = $a[ $orderby ] <=> $b[ $orderby ];

		return $order === 'asc' ? $result : -$result;
	}

	/**
	 * Returns an associative array containing the bulk action
	 *
	 * @return array
	 */
	public function get_bulk_actions() {
		$actions = [
			'bulk-delete' => _x( 'Delete Messages Older Than 1 Year', 'Bulk Action', Plugin::$text_domain ),
		];
		return $actions;
	}

	/**
	 * Handle bulk actions.
	 *
	 * Optional. You can handle your bulk actions anywhere or anyhow you prefer.
	 *
	 * @see $this->prepare_items()
	 */
	public function process_bulk_action() {

		$method = \strtolower( $_SERVER['REQUEST_METHOD'] );
		$action = $this->current_action();
		if ( $method === 'post' ) {
			if ( $action === 'bulk-delete' ) {
				if ( \current_user_can( 'manage_options' ) ) {
					if ( \wp_verify_nonce( $_POST['_wpnonce'], 'bulk-' . $this->_args['plural'] ) ) {
						$guids = \array_map( '\sanitize_text_field', $_POST['guid'] ?? [] );
						$deleted = 0;
						foreach ( $guids as $guid ) {
							$email = UAPI::main_email_account()->email === $this->_args['email'] ? '' : $this->_args['email'];
							$res = UAPI::delete_messages( $email, $guid );
							$deleted++;
						}
						Plugin::delete_transients();
?>
						<div class="notice notice-success is-dismissible">
							<p><?php \printf(
								\_nx(
									'%1$d mailbox reduced.',
									'%1$d mailboxes reduced.',
									\intval( $deleted ),
									'Action result notice, %1$d = number of mailboxes',
									Plugin::$text_domain
								),
								\intval( $deleted ),
							); ?></p>
						</div>
<?php
					}
				}
			}
		} else {
			if ( $action === 'row-delete' ) {
				if ( \current_user_can( 'manage_options' ) ) {
					if ( \wp_verify_nonce( $_GET['_wpnonce'], 'row-delete' ) ) {
						$email = UAPI::main_email_account()->email === $this->_args['email'] ? '' : $this->_args['email'];
						$guid  = \sanitize_text_field( $_GET['guid'] ?? '' );
						$box   = \sanitize_text_field( $_GET['box' ] ?? '' );
						$res = UAPI::delete_messages( $email, $guid, 'savedbefore 1h' );
//						if ( ! $res->has_errors() ) {
							Plugin::delete_transients();
?>
								<div class="notice notice-success is-dismissible">
									<p><?php \printf(
										\_x( 'All messages older than 1 hour in mailbox %1$s for %2$s deleted.', 'Notice success', Plugin::$text_domain ),
										$box,
										Plugin::email_to_utf8( $email ),
									); ?></p>
								</div>
<?php
//						}
					} else {
						\wp_die( \_x( 'Intent mismatch. Go back, refresh page and try again.', 'Die message', Plugin::$text_domain ) );
					}
				}
			}
		}
	}

	private static function display_mailbox( $box ) {
		$parts = \explode( '.', \trim( $box ), 2 );
		$box = \count( $parts ) > 1 && $parts[0] === 'INBOX' ? $parts[1] : $box;
		$box = \ucfirst( \strtolower( $box ) );
		$box_names = [
			'Inbox'   => \_x( 'Inbox',           'Box name', Plugin::$text_domain ),
			'Archive' => \_x( 'Archive',         'Box name', Plugin::$text_domain ),
			'Drafts'  => \_x( 'Draft',           'Box name', Plugin::$text_domain ),
			'Sent'    => \_x( 'Sent',            'Box name', Plugin::$text_domain ),
			'Spam'    => \_x( 'Spam',            'Box name', Plugin::$text_domain ),
			'Trash'   => \_x( 'Trash',           'Box name', Plugin::$text_domain ),
			'Starred' => \_x( 'Stjernemerket',   'Box name', Plugin::$text_domain ),
		];
		return \array_key_exists( $box, $box_names ) ? $box_names[ $box ] : $box;
	}
}
