<?php
namespace WebFacing\cPanel\Email;

/**
 * Exit if accessed directly
 */
if ( ! \class_exists( '\WP' ) ) {
	exit;
}

class List_Table extends \WP_List_Table {

	/**
	 * Prepare the items for the table to process
	 *
	 * @return never
	 */
	public function prepare_items() {

		$this->process_bulk_action();

		$columns  = $this->get_columns();
		$hidden   = $this->get_hidden_columns();
		$sortable = $this->get_sortable_columns();
		$data	  = $this->table_data();
		$cur_page = $this->get_pagenum();
		$per_page = 50;

		if ( \array_key_exists( 'email', $this->_args ) ) {
			$data = \array_filter( $data, function( $row ) {
				return $row['email'] === $this->_args['email'];
			} );
		}

		\usort( $data, [ $this, 'sort_data' ] );
		$total_items  = \count( $data );

		$this->set_pagination_args( [
			'total_items' => $total_items,
			'per_page'    => $per_page
		] );

		$data = \array_slice( $data, ( $cur_page - 1 ) * $per_page, $per_page );

		$this->_column_headers = [ $columns, $hidden, $sortable ];
		$this->items = $data;
	}

	/**
	 * Override the parent columns method. Defines the columns to use in your listing table
	 *
	 * @return array
	 */
	public function get_columns(): array {
		$columns = [
			'cb'        => '<input type="checkbox" />',
			'domain'    => \_x( 'Domain', 'Column Header', Plugin::$text_domain ),
			'email'	    => \_x( 'Email', 'Column Header', Plugin::$text_domain ),
			'type'      => \_x( 'Type', 'Column Header', Plugin::$text_domain ),
			'diskused'  => \_x( 'Used', 'Column Header', Plugin::$text_domain ),
			'diskused_num' => 'DUN',
			'diskquota' => \_x( 'Quota', 'Column Header', Plugin::$text_domain ),
			'dest'      => \_x( 'Destination', 'Column Header', Plugin::$text_domain ),
		];

		return $columns;
	}

	/**
	 * Define which columns are hidden
	 *
	 * @return array
	 */
	public function get_hidden_columns(): array {
		return [ 'diskused_num' ];
	}

	/**
	 * Define the sortable columns
	 *
	 * @return array
	 */
	public function get_sortable_columns(): array {
		return [
			'domain'    => [ 'domain', false ],
			'email'     => [ 'email',  false ],
			'type'      => [ 'type',  false ],
			'diskused'  => [ 'diskused_num',  false ],
		];
	}

	/**
	 * Get the table data
	 *
	 * @return array
	 */
	private function table_data(): array {
		$data = [];
		$account = UAPI::main_email_account();
		$session = UAPI::create_main_webmail_session();
		$data[] = [
			'domain'       => $account->domain,
			'email'        => $account->email,
			'type'         => '_account',
			'diskused'     => \size_format( $account->_diskused ),
			'diskused_num' => $account->_diskused,
			'diskquota'    => $account->diskquota,
			'dest'         => '<form action="https://' . ( $session->hostname ?? Plugin::$host_name ) . ':2096' . $session->token . '/login' . '" method="post" class="webmail" style="display: inline;"><input type="hidden" name="session" value="' . $session->session . '" /><button type="submit" formtarget="' . $account->email . '" class="button button-secondary" title="' . _x( 'If disabled, please refresh this page.', 'Webmail button tooltip.', Plugin::$text_domain ) . '">cPanel® Webmail</button></form> <small style="vertical-align: bottom;">' . \__( '(opens in a new tab)', Plugin::$text_domain ) . '</small>',
		];

		foreach ( (array)UAPI::email_forwarders() as $account ) {
			$data[] = [
				'domain'    => \explode( '@', $account->dest )[1],
				'email'     => $account->html_dest,
				'type'      => 'forwarder',
				'diskused'  => '&mdash;',
				'diskused_num' => 0,
				'diskquota' => '&mdash;',
				'dest'      => \trim( $account->html_forward, '"' ),
			];
		}

		foreach ( (array)UAPI::email_accounts() as $account ) {
			$session = UAPI::create_webmail_session( $account->email );
			$data[] = [
				'domain'    => $account->domain,
				'email'     => $account->email,
				'type'      => 'account',
				'diskused'  => \size_format( $account->_diskused ),
				'diskused_num' => $account->_diskused,
				'diskquota' => $account->_diskquota ? \size_format( $account->_diskquota ) : $account->diskquota,
				'dest'      => '<form action="https://' . ( $session->hostname ?? Plugin::$host_name ) . ':2096' . $session->token . '/login' . '" method="post" class="webmail" style="display: inline;"><input type="hidden" name="session" value="' . $session->session . '" /><button type="submit" formtarget="' . $account->email . '" class="button button-secondary" title="' . _x( 'If disabled, please refresh this page.', 'Webmail button tooltip.', Plugin::$text_domain ) . '">cPanel® Webmail</button></form> <small style="vertical-align: bottom;">' . \__( '(opens in a new tab)', Plugin::$text_domain ) . '</small>',
			];
		}

		foreach ( (array)UAPI::mail_domains() as $domain ) {
			$responders = (array)UAPI::email_responders( $domain );
			foreach ( $responders as $responder ) {
				$data[] = [
					'domain'    => $domain,
					'email'     => $responder->email,
					'type'      => 'responder',
					'diskused'  => '&mdash;',
					'diskused_num' => 0,
					'diskquota' => '&mdash;',
					'dest'      => \esc_html( '↩️ ' . $responder->subject ),
				];
			}
			$data[] = [
				'domain'    => $domain,
				'email'     => '*@' . $domain,
				'type'      => 'default',
				'diskused'  => '&mdash;',
				'diskused_num' => 0,
				'diskquota' => '&mdash;',
				'dest'      => \esc_html( \trim( UAPI::default_address( $domain ), ' "' ) ),
			];
		}

		return $data;
	}

	/**
	 * Render the bulk edit checkbox
	 *
	 * @param array $item
	 *
	 * @return string
	 */
	public function column_cb( $item ) {
		return $item['type'] !== '_account' &&
			( $item['type'] !== 'default' || ! \str_contains( $item[ 'dest' ], ':fail:' ) ) ?
			\sprintf(
				'<input type="checkbox" name="type[%1$s]" value="%2$s" /><input type="hidden" name="dest[%1$s]" value="%3$s" />',
				\esc_attr( $item['email'] ),
				\esc_attr( $item['type' ] ),
				\esc_attr( $item['dest' ] )
			) :
			''
		;
	}

	public function column_email( $item ) {
		$page = \esc_attr( $_REQUEST['page'] );
		if ( $item['type'] === '_account' ) {
			$actions = [
				'postboxes' => '<a href="' . \add_query_arg( [
					'page'     => Plugin::$pf . 'boxes',
					'email'    => $item['email']
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Show mailboxes','Link Table Row Action', Plugin::$text_domain ),
				'row-send'   => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'email'    => $item['email'],
					'action'   => 'row-send',
					'_wpnonce' => \wp_create_nonce( 'row-send' ),
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Get Instructions', 'List Table Row Action', Plugin::$text_domain ),
			];
		} elseif ( $item['type'] === 'forwarder' ) {
			$actions = [
				'row-delete' => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'email'    => $item['email'],
					'dest'     => $item['dest'],
					'action'   => 'row-delete',
					'_wpnonce' => \wp_create_nonce( 'row-delete' )
				], \admin_url( 'admin.php' ) ) . '">' . __( 'Delete', Plugin::$text_domain ),
			];
		} elseif ( $item['type'] === 'default' ) {
			$actions = \str_contains( $item[ 'dest' ], ':fail:' ) ? [] : [
				'row-delete' => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'domain'   => $item['domain'],
					'action'   => 'row-delete',
					'_wpnonce' => \wp_create_nonce( 'row-delete' )
				], \admin_url( 'admin.php' ) ) . '">' . __( 'Remove', Plugin::$text_domain ),
			];
		} elseif ( $item['type'] === 'account' ) {
			$actions = [
				'postboxes' => '<a href="' . \add_query_arg( [
					'page'     => Plugin::$pf . 'boxes',
					'email'    => $item['email']
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Show mailboxes','Link Table Row Action', Plugin::$text_domain ),
				'row-password' => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'email'    => $item['email'],
					'action'   => 'row-password',
					'_wpnonce' => \wp_create_nonce( 'row-password' ),
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Change Password', 'List Table Row Action', Plugin::$text_domain ),
				'row-quota' => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'email'    => $item['email'],
					'action'   => 'row-quota',
					'_wpnonce' => \wp_create_nonce( 'row-quota' ),
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Change Quota', 'List Table Row Action', Plugin::$text_domain ),
				'row-send'     => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'type'     => $item['type'],
					'email'    => $item['email'],
					'action'   => 'row-send',
					'_wpnonce' => \wp_create_nonce( 'row-send' ),
				], \admin_url( 'admin.php' ) ) . '">' . _x( 'Get Instructions', 'List Table Row Action', Plugin::$text_domain ),
				'row-delete'   => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'email'    => $item['email'],
					'type'     => $item['type'],
					'action'   => 'row-delete',
					'_wpnonce' => \wp_create_nonce( 'row-delete' )
				], \admin_url( 'admin.php' ) ) . '">' . __( 'Delete' ),
			];
		} elseif ( $item['type'] === 'responder' ) {
			$actions = [
				'row-edit'   => '<a href="' . \add_query_arg( [
					'page'     => Plugin::$pf . 'new-email',
					'email'    => $item['email'],
					'type'     => $item['type'],
					'action'   => 'row-edit',
				], \admin_url( 'admin.php' ) ) . '#edit-responder">' . __( 'Edit' ),
				'row-delete'   => '<a href="' . \add_query_arg( [
					'page'     => $page,
					'email'    => $item['email'],
					'type'     => $item['type'],
					'action'   => 'row-delete',
					'_wpnonce' => \wp_create_nonce( 'row-delete' )
				], \admin_url( 'admin.php' ) ) . '">' . __( 'Delete' ),
			];
		}
		$format = $item['type'] === 'default' ? '%1$s' : '<a href="mailto:%2$s" title="%3$s %1$s.">%1$s</a><br />%4$s';
		$email  = Plugin::email_to_utf8( $item['email'] );
		return Plugin::email_mx_self( $item['domain'] ) ?
			\sprintf( $format, $email, $item['email'], \_x( 'Send email to', 'Link title text', Plugin::$text_domain ), $this->row_actions( $actions ) ) :
			\sprintf( '<abbr style="color: orangered;" title="%2$s">%1$s</abbr><br /> %3$s', $email, _x( 'Email not hosted on this server - this account will not receive remote incoming emails.', 'Warning title text', Plugin::$text_domain ), $this->row_actions( $actions ) )
		;
	}

	public function column_domain( $item ) {
		$domain = \idn_to_utf8( $item['domain'] );
		return Plugin::email_mx_self( $item['domain'] ) ? $domain : '<abbr style="color: orangered;" title="' . _x( 'Email not hosted on this server - this account will not receive remote incoming emails.', 'Warning title text', Plugin::$text_domain ) . '">' . $domain . '</abbr>';
	}

	public function column_diskused( $item ) {
		return '<a href="'. \add_query_arg( [ 'page'  => Plugin::$pf . 'boxes', 'email' => $item['email'] ], \admin_url( 'admin.php' ) ) . '" title="' . _x( 'View mailboxes.','Link title text', Plugin::$text_domain ) . '">' . $item[ 'diskused' ] . '</a>';
	}

	/**
	 * Define what data to show on each column of the table
	 *
	 * @param  array  $item	       - Data
	 * @param  string $column_name - Current column name
	 *
	 * @return mixed
	 */
	public function column_default( $item, $column_name ) {
		switch ( $column_name ) {
			case 'domain':
			case 'diskused':
			case 'diskquota':
				return $item[ $column_name ];
			case 'type':
				return self::display_type( $item[ $column_name ] );
			case 'email':
			case 'dest':
				return \in_array( $item['type'], [ 'forwarder', 'default' ] ) && \str_contains( $item[ $column_name ], '@' ) ? '↪️ ' . $item[ $column_name ] : ( \str_replace( [ ':fail:',':blackhole:' ], [ '🚫', '⚫' ], $item[ $column_name ] ) );

			default:
				return \print_r( $item, true ) ;
		}
	}

	/**
	 * Allows you to sort the data by the variables set in the $_GET
	 *
	 * @return mixed
	 */
	private function sort_data( $a, $b ) {
		// Set defaults
		$orderby  = 'domain';
		$order    = 'asc';
		$main_domain = UAPI::main_domain();

		// If orderby is set, use this as the sort column
		$orderby  = \sanitize_text_field( $_GET['orderby'] ?? $orderby );
//		$orderby  = $orderby === 'diskused' ? 'diskused_num' : $orderby;

		// If order is set use this as the order
		$order    = \sanitize_text_field( $_GET['order'] ?? $order );

		$a_domain = \intval( $orderby === 'domain' && $main_domain !== $a['domain'] );
		$b_domain = \intval( $orderby === 'domain' && $main_domain !== $b['domain'] );

		$result = $a_domain . $a[ $orderby ] <=> $b_domain . $b[ $orderby ];

		return $order === 'asc' ? $result : -$result;
	}

	/**
	 * Returns an associative array containing the bulk action
	 *
	 * @return array
	 */
	public function get_bulk_actions() {
		$actions = [
			'bulk-delete' => _x( 'Delete Address',          'List Table Bulk Action', Plugin::$text_domain ),
			'bulk-send'   => _x( 'Send Setup Instructions', 'List Table Bulk Action', Plugin::$text_domain ),
		];
		return $actions;
	}

	/**
	 * Handle bulk actions.
	 *
	 * Optional. You can handle your bulk actions anywhere or anyhow you prefer.
	 *
	 * @see $this->prepare_items()
	 */
	public function process_bulk_action() {

		$method = \strtolower( $_SERVER['REQUEST_METHOD'] );
		$action = $this->current_action();
		if ( $method === 'get' ) {
			if ( $action === 'row-password' ) {
				$email = \sanitize_text_field( $_GET['email'] );
?>
				<script>
					var data = {
						'action'  : '<?php echo Plugin::$pf; ?>password',
						'email'   : '<?php echo $email; ?>',
						'password': prompt( '<?php echo __( 'New password for:', Plugin::$text_domain ), ' ', $email; ?>' )
					}
					jQuery.post( ajaxurl, data, function( response) {
						alert( response );
					} );
				</script>
<?php
			} elseif ( $action === 'row-quota' ) {
				$email = \sanitize_text_field( $_GET['email'] );
?>
				<script>
					var data = {
						'action'  : '<?php echo Plugin::$pf; ?>quota',
						'email'   : '<?php echo $email; ?>',
						'quota': prompt( '<?php echo __( 'New quota for:', Plugin::$text_domain ), ' ', $email; ?> (MB)' )
					}
					jQuery.post( ajaxurl, data, function( response) {
						alert( response );
						location.reload();
					} );
				</script>
<?php
			} elseif ( 'row-delete' === $action ) {
				if ( \current_user_can( 'manage_options' ) ) {
					if ( \wp_verify_nonce( $_GET['_wpnonce'], 'row-delete' ) ) {
						$type  = \sanitize_text_field( $_GET['type' ] );
						$email = \sanitize_text_field( $_GET['email'] );
						$dest  = \sanitize_text_field( $_GET['dest' ] );
						if ( $type === 'forwarder' ) {
							$res = UAPI::delete_forward( $email, $dest );
							if ( ! $res->has_errors() ) {
?>
								<div class="notice notice-success is-dismissible">
									<p><?php \printf( \_x( 'Email forwarder for %1$s removed.', '%1$s = email', Plugin::$text_domain ), Plugin::email_to_utf8( $email ) ); ?></p>
								</div>
<?php
							}
						} elseif ( $type === 'account' ) {
							$res = UAPI::delete_account( $email );
							if ( ! $res->has_errors() ) {
?>
								<div class="notice notice-success is-dismissible">
									<p><?php \printf( \_ex( 'Email account %1$s deleted.', '%1$s = email', Plugin::$text_domain ), Plugin::email_to_utf8( $email ) ); ?></p>
								</div>
<?php
							}
						} elseif ( $type === 'default' ) {
							$domain = \sanitize_text_field( $_GET['domain'] );
							$res = UAPI::set_default_fail( $domain );
							if ( ! $res->has_errors() ) {
?>
								<div class="notice notice-success is-dismissible">
									<p><?php \printf( \_x( 'Default address reset for %1$s.', '%1$s = domain', Plugin::$text_domain ), \idn_to_utf8( $domain ) ); ?></p>
								</div>
<?php
							}
						} elseif ( $type === 'responder' ) {
							$res = UAPI::delete_responder( $email );
							if ( ! $res->has_errors() ) {
?>
								<div class="notice notice-success is-dismissible">
									<p><?php \printf( \_x( 'Responder deleted for %1$s.', '%1$s = email', Plugin::$text_domain ), Plugin::email_to_utf8( $email ) ); ?></p>
								</div>
<?php
							}
						}
						Plugin::delete_transients();
					} else {
						\wp_die( \_x( 'Intent mismatch. Go back, refresh page and try again.', 'Die message', Plugin::$text_domain ) );
					}
				}
			} elseif ( 'row-send' === $this->current_action() ) {
				if ( \current_user_can( 'manage_options' ) ) {
					$type  = \sanitize_text_field( $_GET['type' ] );
					if ( \in_array( $type, [ '_account', 'account' ], true ) ) {
						$email = \sanitize_text_field( $_GET['email'] );
						$to    = \wp_get_current_user()->user_email;
						$res   = UAPI::send_settings( $email, $to );
						if ( ! $res->has_errors() ) {
?>
							<div class="notice notice-success is-dismissible">
								<p>Email account client instructions for account <?php echo \esc_attr( $email ); ?> sent to <?php echo \esc_attr( $to ) ;?>.</p>
							</div>
<?php
						}
					}
				}
			}
		} elseif ( $method === 'post' ) {
			if ( 'bulk-delete' === $this->current_action() ) {
				if ( \current_user_can( 'manage_options' ) ) {
					if ( \wp_verify_nonce( $_POST['_wpnonce'], 'bulk-' . $this->_args['plural'] ) ) {
						$types = \array_map( '\sanitize_text_field', $_POST['type'] );
						$dests = \array_map( '\sanitize_text_field', $_POST['dest'] );
						$deleted = 0;
						foreach ( $types as $email => $type ) {
							if ( $type === 'forwarder' ) {
								$res = UAPI::delete_forward( $email, $dests[ $email ] );
								if ( ! $res->has_errors() ) {
									$deleted++;
								}
							} elseif ( $type === 'account' ) {
								$res = UAPI::delete_account( $email );
								if ( ! $res->has_errors() ) {
									$deleted++;
								}
							} elseif ( $type === 'default' ) {
								$domain = \explode( '@', $email )[1];
								$res = UAPI::set_default_fail( $domain );
								if ( ! $res->has_errors() ) {
									$deleted++;
								}
							} elseif ( $type === 'responder' ) {
								$res = UAPI::delete_responder( $email );
								if ( ! $res->has_errors() ) {
									$deleted++;
								}
							}
						}
						Plugin::delete_transients();
?>
						<div class="notice notice-warning is-dismissible">
							<p><?php \printf(
								\_nx(
									'%1$d email address type %2$s deleted.',
									'%1$d email addresses type %2$s deleted.',
									\intval( $deleted ),
									'Action result notice, %1$d = number of emails, %2$s = types',
									Plugin::$text_domain
									),
								\intval( $deleted ),
								\implode( ', ', $types )
							); ?></p>
						</div>
<?php
					} else {
						\wp_die( \_x( 'Intent mismatch. Go back, refresh page and try again.', 'Die message', Plugin::$text_domain ) );
					}
				} else {
?>
					<div class="notice notice-error is-dismissible">
						<p><?php \_ex( 'Sorry, you are not allowed to delete email adresses.', 'Notice error message', Plugin::$text_domain ); ?></p>
					</div>
<?php
				}
			} elseif ( 'bulk-send' === $this->current_action() ) {
				if ( \current_user_can( 'manage_options' ) ) {
					if ( \wp_verify_nonce( $_GET['_wpnonce'], 'bulk-' . $this->_args['plural'] ) ) {
						$types = \array_map( '\sanitize_text_field', $_GET['type'] );
						$to = \wp_get_current_user()->user_email;
						$sent = 0;
						foreach ( $types as $email => $type ) {
							if ( \in_array( $type, [ '_account', 'account' ], true ) ) {
								$res = UAPI::send_settings( $email, $to );
								if ( ! $res->has_errors() ) {
									$sent++;
								}
							}
						}
?>
						<div class="notice notice-success is-dismissible">
							<p><?php \printf(
								\_nx(
									'%1$d email account client instructions sent to %2$s.',
									'%1$d email accounts client instructions sent to %2$s.',
									\intval( $sent ),
									'Notice success message, %1$d = number of accounts, %2$s = destination email',
									Plugin::$text_domain
								),
								\intval( $sent ),
								\esc_attr( Plugin::email_to_utf8( $to ) ) );
							?></p>
						</div>
<?php
					} else {
						\wp_die( \_x( 'Intent mismatch. Go back, refresh and try again.', 'Die message', Plugin::$text_domain ) );
					}
				} else {
?>
					<div class="notice notice-error is-dismissible">
						<p><?php \_ex( 'Sorry, you are not allowed to send settings.', 'Notice error message', Plugin::$text_domain ); ?></p>
					</div>
<?php
				}
			}
		}
	}

	private static function display_type( $type ) {
		$type_names = [
			'_account'   => \_x( 'Main Account', 'Email entry type', Plugin::$text_domain ),
			 'account'   => \_x(      'Account', 'Email entry type', Plugin::$text_domain ),
			 'forwarder' => \_x( 'Forwarder',    'Email entry type', Plugin::$text_domain ),
			 'default'   => \_x( 'Default',      'Email entry type', Plugin::$text_domain ),
			 'responder' => \_x( 'Responder',    'Email entry type', Plugin::$text_domain ),
		];
		return \array_key_exists( $type, $type_names ) ? $type_names[ $type ] : \ucfirst( $type );
	}
}
